import hashlib
import logging
import tempfile
from enum import Enum
from itertools import product
from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, validator
from ruamel.yaml import YAMLError

from .common import HarrierProblem, PathMatch, yaml
from .extensions import Extensions

logger = logging.getLogger('harrier.config')
CONFIG_FILE_TRIES = 'harrier', 'config', '_config'
CONFIG_FILE_TRIES = [Path(f'{name}.{ext}') for name, ext in product(CONFIG_FILE_TRIES, ['yml', 'yaml'])]


class Mode(str, Enum):
    development = 'development'
    production = 'production'


class WebpackConfig(BaseModel):
    cli: Path = 'node_modules/.bin/webpack-cli'
    entry: Path = 'js/index.js'
    output_path: Path = 'theme'
    dev_output_filename: Optional[str] = 'main.js'
    prod_output_filename: Optional[str] = 'main.[hash].js'
    config: Path = None
    run: bool = True

    class Config:
        validate_all = True


class Config(BaseModel):
    source_dir: Path
    mode: Mode = Mode.production
    pages_dir: Path = 'pages'
    extensions: Extensions = 'extensions.py'
    theme_dir: Path = 'theme'
    data_dir: Path = 'data'

    dist_dir: Path = 'dist'
    dist_dir_sass: Path = 'theme'
    dist_dir_assets: Path = 'theme/assets'
    tmp_dir: Path = None

    download: Dict[str, Any] = {}
    download_aliases: Dict[str, str] = {}

    default_template: Optional[str] = None
    defaults: Dict[PathMatch, Dict[str, Any]] = {}
    ignore: List[PathMatch] = []

    webpack: WebpackConfig = WebpackConfig()

    @validator('source_dir')
    def resolve_source_dir(cls, v):
        return v.resolve(strict=True)

    @validator('pages_dir', 'theme_dir', 'data_dir')
    def resolve_relative_paths(cls, v, values, **kwargs):
        return (values['source_dir'] / v).resolve()

    @validator('pages_dir')
    def is_dir(cls, v, field, **kwargs):
        if not v.exists():
            raise ValueError(f'{field.name} directory "{v}" does not exist')
        elif not v.is_dir():
            raise ValueError(f'{field.name} "{v}" is not a directory')
        else:
            return v

    @validator('dist_dir')
    def check_dist_dir(cls, p, values, **kwargs):
        if not p.is_absolute():
            p = (values['source_dir'] / p).resolve()
        if not p.parent.exists():
            raise ValueError(f'dist_dir "{p}" parent directory does not exist')
        elif p.exists() and not p.is_dir():
            raise ValueError(f'dist_dir "{p}" is not a directory')
        else:
            return p

    @validator('default_template')
    def theme_templates(cls, v, values, **kwargs):
        templates_dir = values['theme_dir'] / 'templates'
        if not v or templates_dir.is_dir():
            return v
        else:
            raise ValueError(f'default-template set but template directory "{templates_dir}" does not exist')

    @validator('extensions', pre=True)
    def validate_extensions(cls, v, values, **kwargs):
        p = values['source_dir'] / v
        if p.exists() and not p.is_file():
            raise ValueError(f'"extensions" should be a python file, not directory')
        else:
            return p

    @validator('webpack')
    def validate_webpack(cls, v, *, values, **kwargs):
        webpack: WebpackConfig = v
        if not webpack.run:
            return webpack

        if {'source_dir', 'theme_dir', 'source_dir', 'dist_dir'} - values.keys():
            # some values are missing, can't validate properly
            return webpack

        if not webpack.cli.is_absolute():
            webpack.cli = values['source_dir'] / webpack.cli

        if not webpack.cli.exists():
            logger.warning('webpack cli path "%s" does not exist, not running webpack', webpack.cli)
            webpack.run = False

        webpack.entry = values['theme_dir'] / webpack.entry
        if not webpack.entry.exists() and webpack.run:
            logger.warning('webpack entry point "%s" does not exist, not running webpack', webpack.entry)
            webpack.run = False

        if webpack.config:
            webpack.config = values['source_dir'] / webpack.config
            if not webpack.config.exists():
                raise ValueError(f'webpack config set but does not exist "{webpack.config}", not running webpack')

        webpack.output_path = values['dist_dir'] / webpack.output_path
        return webpack

    def get_tmp_dir(self) -> Path:
        if self.tmp_dir:
            return self.tmp_dir
        else:
            path_hash = hashlib.md5(b'%s' % self.source_dir).hexdigest()
            return Path(tempfile.gettempdir()) / f'harrier-{path_hash}'

    class Config:
        allow_extra = True
        validate_all = True


def load_config_file(config_path: Path):
    try:
        raw_config = yaml.load(config_path.read_text()) or {}
    except YAMLError as e:
        logger.error('%s: %s', e.__class__.__name__, e)
        raise HarrierProblem(f'error loading "{config_path}"') from e
    raw_config.setdefault('source_dir', config_path.parent)
    return raw_config


def get_config(path) -> Config:
    config_path = Path(path)
    if config_path.is_file():
        config = load_config_file(config_path)
    else:
        try:
            config_path = next(config_path / f for f in CONFIG_FILE_TRIES if (config_path / f).exists())
        except StopIteration:
            config = {'source_dir': config_path}
        else:
            config = load_config_file(config_path)

    return Config(**config)



import logging
import sys
import traceback

import click
from grablib.common import GrablibError
from pydantic import ValidationError

from . import main
from .common import HarrierProblem, setup_logging
from .config import Mode
from .version import VERSION

steps_help = 'Build steps to run, multiple values allowed, default: all.'
dev_help = 'Whether to build in development or production mode, default: production.'
verbose_help = 'Enable verbose output.'
logger = logging.getLogger('harrier')


@click.group()
@click.version_option(VERSION, '-V', '--version', prog_name='harrier')
def cli():
    """
    harrier - yet another static site builder
    """
    pass


@cli.command()
@click.argument('path', type=click.Path(exists=True), required=False, default='.')
@click.option('--steps', '-s', multiple=True, type=click.Choice(main.ALL_STEPS), help=steps_help)
@click.option('-d/-p', '--dev/--prod', 'dev_mode', default=None, help=dev_help)
@click.option('-v/-q', '--verbose/--quiet', 'verbose', default=None, help=verbose_help)
def build(path, dev_mode, steps, verbose):
    """
    build the site
    """
    setup_logging(verbose)
    mode = None
    if dev_mode is not None:
        mode = Mode.development if dev_mode else Mode.production

    try:
        main.build(path, set(steps), mode)
    except (HarrierProblem, ValidationError, GrablibError) as e:
        msg = 'Error: {}'
        if not verbose:
            msg += '\n\nUse "--verbose" for more details'
        logger.debug(traceback.format_exc())
        logger.error(msg.format(e))
        sys.exit(2)


@cli.command()
@click.argument('path', type=click.Path(exists=True), required=False, default='.')
@click.option('-p', '--port', default=8000, type=int, help='port to use for dev server.')
@click.option('-v/-q', '--verbose/--quiet', 'verbose', default=None, help=verbose_help)
def dev(path, port, verbose):
    """
    Serve the site while watching for file changes and rebuilding upon changes.
    """
    setup_logging(verbose, dev=True)
    try:
        main.dev(path, port)
    except (HarrierProblem, ValidationError, GrablibError) as e:
        msg = 'Error: {}'
        if not verbose:
            msg += '\n\nUse "--verbose" for more details'
        logger.debug(traceback.format_exc())
        logger.error(msg.format(e))
        sys.exit(2)



import asyncio
import json
import logging
import os
import re
import shutil
import subprocess
from functools import partial
from time import time

from grablib.build import SassGenerator, insert_hash
from grablib.common import GrablibError
from grablib.download import Downloader
from pygments.formatters.html import HtmlFormatter

from .common import HarrierProblem, log_complete
from .config import Config, Mode

logger = logging.getLogger('harrier.assets')


def run_grablib(config: Config):
    start = time()
    download_root = config.theme_dir / 'libs'
    log_msg = False
    if config.download:
        logger.debug('running grablib download...')
        download = Downloader(
            download_root=download_root,
            download=config.download,
            aliases=config.download_aliases,
            lock=config.theme_dir / '.grablib.lock',
        )
        download()
        log_msg = True

    sass_dir = config.theme_dir / 'sass'
    count = 0
    if sass_dir.is_dir():
        output_dir = config.dist_dir / config.dist_dir_sass
        output_dir.relative_to(config.dist_dir)

        # this prevents SassGenerator from throwing an error because this directory exists
        out_dir_src = output_dir / '.src'
        out_dir_src.is_dir() and shutil.rmtree(out_dir_src)

        theme_files = find_theme_files(config)
        custom_functions = {
            'resolve_url': partial(resolve_sass_url, theme_files),
        }

        sass_gen = SassGenerator(
            input_dir=sass_dir,
            output_dir=output_dir,
            download_root=download_root,
            debug=config.mode == Mode.development,
            apply_hash=config.mode == Mode.production,
            custom_functions=custom_functions,
            extra_importers=[(0, pygments_importer)]
        )
        try:
            sass_gen()
        except GrablibError as e:
            raise HarrierProblem('error generating sass') from e
        log_msg = True
        count = sass_gen._files_generated

    log_msg and log_complete(start, 'sass built', count)
    return count


PYGMENTS_PREFIX = 'pygments/'


def pygments_importer(path: str):
    if not path.startswith(PYGMENTS_PREFIX):
        return
    style_name = path[len(PYGMENTS_PREFIX):]
    formatter = HtmlFormatter(style=style_name)
    return [(f'pygments/{style_name}.css', formatter.get_style_defs('.hi'))]


def copy_assets(config: Config):
    start = time()
    in_dir = config.theme_dir / 'assets'
    if not in_dir.is_dir():
        return
    out_dir = config.dist_dir / config.dist_dir_assets
    out_dir.relative_to(config.dist_dir)
    copied = 0
    for in_path in in_dir.glob('**/*'):
        out_path = out_dir / in_path.relative_to(in_dir)
        if config.mode == Mode.production:
            out_path = insert_hash(out_path, in_path.read_bytes())
        out_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(in_path, out_path)
        copied += 1
    logger.debug('copied %d theme assets from "%s" to "%s"',
                 copied, in_dir.relative_to(config.source_dir), out_dir.relative_to(config.dist_dir))

    copied and log_complete(start, 'theme assets copied', copied)
    return copied


def assets_grablib(config: Config):
    copy_assets(config)
    run_grablib(config)


def webpack_configuration(config: Config, watch: bool):
    if not config.webpack or not config.webpack.run:
        return None, None

    wp = config.webpack
    prod = config.mode == Mode.production
    output_filename = wp.prod_output_filename if prod else wp.dev_output_filename
    # ./ is required to satisfy webpack when files are inside the "--context" directory
    args = (
        wp.cli,
        '--context', config.source_dir,
        '--entry', f'./{wp.entry.relative_to(config.source_dir)}',
        '--output-path', wp.output_path,
        output_filename and '--output-filename', output_filename,
        '--devtool', 'source-map',
        '--mode', config.mode.value,
        watch and '--watch',
        prod and '--optimize-minimize',
        wp.config and '--config',
        wp.config and f'./{wp.config.relative_to(config.source_dir)}',
    )
    env = dict(**os.environ, **{
        'NODE_ENV': config.mode.value,
        # 'HARRIER_CONFIG': json.dumps(config.dict())  # TODO
    })
    return [str(a) for a in args if a], env


def run_webpack(config: Config):
    start = time()
    args, env = webpack_configuration(config, False)
    if not args:
        return

    cmd = ' '.join(args)
    kwargs = dict(check=True, cwd=config.source_dir, env=env)
    logger.debug('webpack command "%s"', cmd)
    capture_output = not logger.isEnabledFor(logging.DEBUG)
    if capture_output:
        kwargs.update(stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf8')
        args += '--json',
    try:
        p = subprocess.run(args, **kwargs)
    except subprocess.CalledProcessError as e:
        logger.warning('error running webpack "%s", returncode %s\nstdout: %s\nstderr: %s',
                       cmd, e.returncode, e.output, e.stderr)
        raise HarrierProblem('error running webpack') from e
    else:
        count = 1
        if capture_output:
            try:
                output = json.loads(p.stdout[p.stdout.find('{'):])
            except ValueError:
                # happens when the webpack config script writes to standout including a "{"
                pass
            else:
                count = len(output['assets'])
        log_complete(start, 'webpack built', count)
        return count


async def start_webpack_watch(config: Config):
    args, env = webpack_configuration(config, True)
    if args:
        cmd = ' '.join(args)
        logger.info('running webpack ...')
        logger.debug('webpack command "%s"', cmd)
        return await asyncio.create_subprocess_exec(*args, cwd=config.source_dir, env=env)


def find_theme_files(config: Config):
    check_dirs = (
        config.dist_dir / config.dist_dir_sass,
        config.dist_dir / config.dist_dir_assets,
        config.webpack.output_path
    )
    d = {}
    for dir in check_dirs:
        if dir.is_dir():
            for p in dir.glob('**/*'):
                if p.is_file():
                    rel_path = str(p.relative_to(config.dist_dir))
                    path_name = rel_path
                    if config.mode == Mode.production:
                        path_name = re.sub('\.[a-f0-9]{7,20}\.', '.', rel_path)
                        path_name = re.sub('\.[a-f0-9]{7,20}$', '', path_name)
                    d[path_name] = rel_path
    return d


def resolve_sass_url(theme_files, path):
    # TODO try more things, raise error on failure
    return f'"{theme_files.get(path) or path}"'



import logging.config
import re
from fnmatch import translate
from os.path import normcase
from time import time

import click
from pydantic.validators import str_validator
from ruamel.yaml import YAML

yaml = YAML(typ='safe')
completed_logger = logging.getLogger('harrier.completed')


class HarrierProblem(RuntimeError):
    pass


def log_complete(start, description, items):
    completed_logger.info('%6s %20s %0.3fs', items, description, time() - start)


class PathMatch:
    __slots__ = 'raw', '_regex'

    def __init__(self, s):
        self.raw = s
        self._regex = re.compile(translate(normcase(s)))

    def __call__(self, path: str):
        return self._regex.match(path)

    def __eq__(self, other):
        return self.raw == getattr(other, 'raw', None)

    def __hash__(self):
        return hash(self.raw)

    def __repr__(self):
        return f'<PathMatch {self.raw!r}>'

    @classmethod
    def get_validators(cls):
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls, value):
        return cls(value)


class ColourHandler(logging.Handler):  # pragma: no cover
    formats = {
        logging.DEBUG: {'fg': 'white', 'dim': True},
        logging.INFO: {'fg': 'white', 'dim': True},
        logging.WARN: {'fg': 'yellow'},
    }

    def get_log_format(self, record):
        return self.formats.get(record.levelno, {'fg': 'red'})

    def emit(self, record):
        log_entry = self.format(record)
        click.secho(log_entry, **self.get_log_format(record))


def log_config(verbose: bool, dev) -> dict:
    if verbose is True:
        log_level = 'DEBUG'
    elif verbose is False:
        log_level = 'WARNING'
    else:
        assert verbose is None
        log_level = 'INFO'
    return {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'default': {
                'format': '%(message)s'
            },
            'server': {
                'format': '[%(asctime)s] %(message)s',
                'datefmt': '%H:%M:%S',
            },
        },
        'handlers': {
            'default': {
                'level': log_level,
                'class': 'grablib.common.ClickHandler',
                'formatter': 'default'
            },
            'build': {
                'level': 'DEBUG' if verbose else ('WARNING' if dev else 'INFO'),
                'class': 'grablib.common.ClickHandler',
                'formatter': 'default'
            },
            'grablib': {
                'level': 'INFO' if verbose else 'WARNING',
                'class': 'harrier.common.ColourHandler',
                'formatter': 'default'
            },
            'server_logging': {
                'level': log_level,
                'class': 'aiohttp_devtools.runserver.log_handlers.AuxiliaryHandler',
                'formatter': 'server'
            },
        },
        'loggers': {
            'harrier': {
                'handlers': ['default'],
                'level': log_level,
            },
            'harrier.build': {
                'handlers': ['build'],
                'level': log_level,
                'propagate': False,
            },
            'harrier.assets': {
                'handlers': ['build'],
                'level': log_level,
                'propagate': False,
            },
            'grablib': {
                'handlers': ['grablib'],
                'level': log_level,
            },
            'adev.server.aux': {
                'handlers': ['server_logging'],
                'level': log_level,
            },
        },
    }


def setup_logging(verbose, dev=False):
    config = log_config(verbose, dev)
    logging.config.dictConfig(config)



import asyncio
import logging
import shutil
from concurrent.futures import ProcessPoolExecutor
from enum import Enum
from pathlib import Path
from typing import Optional, Set, Union

import devtools

from .assets import assets_grablib, find_theme_files, run_webpack
from .build import build_pages, render_pages
from .common import completed_logger
from .config import Mode, get_config
from .data import load_data
from .dev import adev
from .extensions import apply_modifiers

logger = logging.getLogger('harrier.main')
StrPath = Union[str, Path]


class BuildSteps(str, Enum):
    clean = 'clean'
    extensions = 'extensions'
    copy_assets = 'copy_assets'
    pages = 'pages'
    data = 'data'
    sass = 'sass'
    webpack = 'webpack'


ALL_STEPS = [m.value for m in BuildSteps.__members__.values()]


def build(path: StrPath, steps: Set[BuildSteps]=None, mode: Optional[Mode]=None):
    completed_logger.info('building site...')
    config = get_config(path)
    if mode:
        config.mode = mode
    logger.debug('Config: %s', devtools.pformat(config.dict()))

    steps = steps or ALL_STEPS
    if BuildSteps.extensions in steps:
        config = apply_modifiers(config, config.extensions.config_modifiers)

    clean = BuildSteps.clean in steps
    _empty_dir(config.dist_dir, clean)
    _empty_dir(config.get_tmp_dir(), clean)

    pages = None
    data_future = None
    with ProcessPoolExecutor() as executor:
        futures = [
            BuildSteps.sass in steps and executor.submit(assets_grablib, config),
            BuildSteps.webpack in steps and executor.submit(run_webpack, config),
        ]

        if BuildSteps.data in steps:
            data_future = executor.submit(load_data, config)

        if BuildSteps.pages in steps:
            pages = build_pages(config)
        # this will raise errors if any of the above went wrong
        [f.result() for f in futures if f]

    som = dict(
        theme_files=find_theme_files(config),
        pages=pages,
        data=data_future and data_future.result(),
        **config.dict(),
    )

    if BuildSteps.extensions in steps:
        som = apply_modifiers(som, config.extensions.som_modifiers)

    if som['pages'] is not None:
        render_pages(config, som)
    return som


def dev(path: StrPath, port: int):
    config = get_config(path)
    config.mode = Mode.development
    logger.debug('Config:\n%s', devtools.pformat(config.dict()))

    _empty_dir(config.dist_dir)
    _empty_dir(config.get_tmp_dir())

    loop = asyncio.get_event_loop()
    return loop.run_until_complete(adev(config, port))


def _empty_dir(d: Path, clean: bool=True):
    if clean and d.exists():
        shutil.rmtree(d)
    d.mkdir(exist_ok=True, parents=True)



import csv
import json
import logging
import re
from pathlib import Path
from time import time

from ruamel.yaml import YAMLError

from .common import HarrierProblem, log_complete, yaml
from .config import Config

logger = logging.getLogger('harrier.data')
csv_dialect = csv.excel
csv_dialect.skipinitialspace = True


def load_data(config: Config):
    start = time()
    d = config.data_dir
    if not d.is_dir():
        return None
    ext_lookup = {
        '.json': read_json,
        '.csv': read_csv,
        '.yaml': read_yaml,
        '.yml': read_yaml,
    }
    data = {}
    count = 0
    for ext in ext_lookup:
        for p in d.glob(f'**/*{ext}'):
            key = re.sub('[-/ ]', '_', str(p.relative_to(d).with_suffix('')))
            key = re.sub('\W', '', key)
            if key in data:
                logger.warning('duplicate data key "%s", ignoring data in "%s", please rename', key, p)
                continue
            logger.debug('reading data from "%s" as "%s"', p, key)
            try:
                data[key] = ext_lookup[p.suffix](p)
            except (ValueError, YAMLError) as e:
                raise HarrierProblem(f'error reading {p} {e.__class__.__name__}: {e}') from e
            count += 1
    log_complete(start, 'data loaded', count)
    return data


def read_json(p: Path):
    with p.open() as f:
        return json.load(f)


def read_csv(p: Path):
    with p.open(newline='') as f:
        reader = csv.DictReader(f, dialect=csv_dialect, restval='other')
        return [dict(r) for r in reader]


def read_yaml(p: Path):
    return yaml.load(p.read_text())



import hashlib
import logging
import os
import re
import shutil
from datetime import datetime
from pathlib import Path
from time import time
from typing import Optional

from jinja2 import Environment, FileSystemLoader, contextfunction
from misaka import HtmlRenderer, Markdown, escape_html
from pydantic import BaseModel, validator
from pygments import highlight
from pygments.formatters import ClassNotFound
from pygments.formatters.html import HtmlFormatter
from pygments.lexers import get_lexer_by_name
from ruamel.yaml import YAMLError

from .common import HarrierProblem, log_complete, yaml
from .config import Config
from .extensions import ExtensionError

FRONT_MATTER_START_REGEX = re.compile(r'---[ \t]*(.*)\n---[ \t]*\n', re.S)
FRONT_MATTER_DIVIDER_REGEX = re.compile(r'\n?^--- ?([.\w_-]+) ?---[ \t]*\n', re.S | re.M)
# extensions where we want to do anything except just copy the file to the output dir
OUTPUT_HTML = {'.html', '.md'}
MAYBE_RENDER = {'.xml'}
URI_NOT_ALLOWED = re.compile(r'[^a-zA-Z0-9_\-/.]')
DATE_REGEX = re.compile(r'(\d{4})-(\d{2})-(\d{2})-?(.*)')
URI_IS_TEMPLATE = re.compile('[{}]')

logger = logging.getLogger('harrier.build')


def build_pages(config: Config):
    start = time()
    pages, files = BuildPages(config).run()
    log_complete(start, 'pages built', files)
    return pages


def render_pages(config: Config, som: dict, build_cache=None):
    start = time()
    cache, files = Renderer(config, som, build_cache).run()
    log_complete(start, 'pages rendered', files)
    return cache


class BuildPages:
    __slots__ = 'config', 'tmp_dir', 'files', 'template_files'

    def __init__(self, config: Config):
        self.config = config
        self.tmp_dir = config.get_tmp_dir()
        self.files = 0
        self.template_files = 0

    def run(self):
        paths = sorted(self.config.pages_dir.glob('**/*'), key=lambda p_: (len(p_.parents), str(p_)))
        pages = {}
        for p in paths:
            if p.is_file():
                try:
                    v = self.prep_file(p)
                except ExtensionError:
                    # these are logged directly
                    raise
                except Exception:
                    logger.exception('%s: error building SOM for page', p)
                    raise
                if v:
                    pages[str(p.relative_to(self.config.pages_dir))] = v
        logger.debug('Built site object model with %d files, %d files to render', self.files, self.template_files)
        return pages, self.files

    def prep_file(self, p):
        path_ref = '/' + os.path.normcase(str(p.relative_to(self.config.pages_dir)))
        if any(path_match(path_ref) for path_match in self.config.ignore):
            return
        html_output = p.suffix in OUTPUT_HTML

        data, apple_template = self.get_page_data(p, html_output, path_ref)

        for path_match, f in self.config.extensions.page_modifiers:
            if path_match(path_ref):
                try:
                    data = f(data, config=self.config)
                except Exception as e:
                    logger.exception('%s error running page extension %s', p, f.__name__)
                    raise ExtensionError(str(e)) from e
                if not isinstance(data, dict):
                    logger.error('%s extension "%s" did not return a dict', p, f.__name__)
                    raise ExtensionError(f'extension "{f.__name__}" did not return a dict')

        fd = FileData(**data)
        final_data = fd.dict(exclude=set() if apple_template else {'template', 'render'})
        final_data['__file__'] = 1

        if apple_template and fd.render:
            if not fd.content_template.parent.exists():
                fd.content_template.parent.mkdir(parents=True)
            fd.content_template.write_text(final_data.pop('content'))
            final_data['content_template'] = str(fd.content_template.relative_to(self.tmp_dir))
        else:
            final_data.pop('content_template')

        # logger.debug('added %s apple_template: %s, outfile %s', p, apple_template, fd.outfile)
        self.files += 1
        if apple_template:
            self.template_files += 1

        return final_data

    def get_page_data(self, p, html_output, path_ref):  # noqa: C901 (ignore complexity)
        name = p.stem if html_output else p.name

        date_match = DATE_REGEX.match(name)
        if date_match:
            *date_args, new_name = date_match.groups()
            created = datetime(*map(int, date_args))
            name = new_name or name
        else:
            created = p.stat().st_mtime

        data = {
            'infile': p,
            'content_template': self.tmp_dir / 'content' / p.relative_to(self.config.pages_dir),
            'template': self.config.default_template,
            'title': name,
            'slug': '' if html_output and p.stem == 'index' else slugify(name),
            'created': created,
        }

        for path_match, defaults in self.config.defaults.items():
            if path_match(path_ref):
                data.update(defaults)

        maybe_render = p.suffix in MAYBE_RENDER
        apple_template = data.get('apply_template', None)
        if apple_template is not False and html_output or maybe_render:
            fm_data, content = parse_front_matter(p.read_text())
            if html_output or fm_data:
                data['content'] = content
                fm_data and data.update(fm_data)
                apple_template = True

        uri = data.get('uri')
        if not uri:
            parents = str(p.parent.relative_to(self.config.pages_dir)).split('/')
            if parents == ['.']:
                data['uri'] = '/' + data['slug']
            else:
                data['uri'] = '/' + '/'.join([slugify(p) for p in parents] + [data['slug']])
        elif URI_IS_TEMPLATE.search(uri):
            try:
                data['uri'] = slugify(uri.format(**data))
            except KeyError as e:
                raise KeyError(f'missing format variable "{e.args[0]}" for "{uri}"')

        if data.get('output', True):
            outfile = self.config.dist_dir / data['uri'][1:]
            if html_output and outfile.suffix != '.html':
                outfile /= 'index.html'
            data['outfile'] = outfile
        return data, apple_template


class HighlighterRenderer(HtmlRenderer):
    def blockcode(self, text, lang):
        try:
            lexer = get_lexer_by_name(lang, stripall=True)
        except ClassNotFound:
            lexer = None

        if lexer:
            formatter = HtmlFormatter(cssclass='hi')
            return highlight(text, lexer, formatter)
        # default
        code = escape_html(text.strip())
        return f'<pre><code>{code}</code></pre>\n'


class Renderer:
    __slots__ = 'config', 'som', 'build_cache', 'md', 'env', 'checked_dirs'

    def __init__(self, config: Config, som: dict, build_cache: dict=None):
        self.config = config
        self.som = som
        self.build_cache = build_cache

        md_renderer = HighlighterRenderer()
        self.md = Markdown(md_renderer, extensions=('fenced-code',))

        template_dirs = [str(self.config.get_tmp_dir()), str(self.config.theme_dir / 'templates')]
        logger.debug('template directories: %s', ', '.join(template_dirs))

        self.env = Environment(loader=FileSystemLoader(template_dirs))
        self.env.filters.update(self.config.extensions.template_filters)

        self.env.globals['url'] = resolve_url
        self.env.globals.update(self.config.extensions.template_functions)

        self.checked_dirs = set()

    def run(self):
        gen, copy = 0, 0
        for p in self.som['pages'].values():
            action = self.render_file(p)
            if action == 1:
                gen += 1
            elif action == 2:
                copy += 1
        logger.debug('generated %d files, copied %d files', gen, copy)
        return self.build_cache, gen + copy

    def render_file(self, p):
        if not p.get('outfile'):
            return
        outfile: Path = p['outfile']
        out_dir = outfile.parent
        if out_dir not in self.checked_dirs:
            # this will raise an exception if somehow outfile is outside dis_dir
            out_dir.relative_to(self.config.dist_dir)
            out_dir.mkdir(exist_ok=True, parents=True)
            self.checked_dirs.add(out_dir)

        infile: Path = p['infile']
        if 'template' in p:
            return self.render_template(p, infile, outfile)
        else:
            return self.copy_file(p, infile, outfile)

    def render_template(self, p: dict, infile: Path, outfile: Path):
        template_file = p['template']
        try:
            if p['render']:
                content_template = self.env.get_template(str(p['content_template']))
                content = content_template.render(page=p, site=self.som)
            else:
                content = p['content']

            content = split_content(content)

            if infile.suffix == '.md':
                if isinstance(content, dict):
                    content = {k: self.md(v) for k, v in content.items()}
                elif isinstance(content, list):
                    content = [self.md(v) for v in content]
                else:
                    # assumes content is a str
                    content = self.md(content)

            if template_file:
                template = self.env.get_template(template_file)
                rendered = template.render(content=content, page=p, site=self.som)
            else:
                rendered = content
            rendered = rendered.rstrip(' \t\r\n') + '\n'
        except Exception:
            logger.exception('%s: error rendering page', infile)
            raise
        else:
            rendered_b = rendered.encode()
            if self.build_cache is not None:
                out_hash = hashlib.md5(rendered_b).digest()
                if self.build_cache.get(infile) == out_hash:
                    # file hasn't changed
                    return
                else:
                    self.build_cache[infile] = out_hash
            outfile.write_bytes(rendered_b)
            return 1

    def copy_file(self, p: dict, infile: Path, outfile: Path):
        if self.build_cache is not None:
            mtime = infile.stat().st_mtime
            if self.build_cache.get(infile) == mtime:
                # file hasn't changed
                return
            else:
                self.build_cache[infile] = mtime
        shutil.copy(infile, outfile)
        return 2


def parse_front_matter(s):
    m = FRONT_MATTER_START_REGEX.match(s)
    if not m:
        return None, s
    try:
        data = yaml.load(m.groups()[0]) or {}
    except YAMLError as e:
        raise HarrierProblem(f'error parsing YAML: {e}') from e
    content = s[m.end():]
    return data, content


def split_content(s):
    m = FRONT_MATTER_DIVIDER_REGEX.search(s)
    if not m:
        return s
    content = []
    name = 'main'
    while True:
        start, end = m.span()
        content.append(
            (name, s[:start])
        )
        name = m.groups()[0]
        s = s[end:]
        m = FRONT_MATTER_DIVIDER_REGEX.search(s)
        if not m:
            break
    content.append(
        (name, s)
    )
    names, values = zip(*content)
    names = set(names[1:])
    if names == {'.'}:
        return list(values)
    elif '.' in names:
        raise HarrierProblem(f'badly constructed multi-part front matter, dividers indicate a mix of list and dict')
    else:
        return {k: v for k, v in content if v}


class FileData(BaseModel):
    infile: Path
    content_template: Path
    title: str
    slug: str
    created: datetime
    uri: str
    template: Optional[str]
    render: bool = True
    outfile: Path = None

    @validator('uri')
    def validate_uri(cls, v):
        if not v.startswith('/'):
            raise ValueError(f'uri must start with a slash: "{v}')
        invalid = URI_NOT_ALLOWED.findall(v)
        if invalid:
            invalid = ', '.join(f'"{inv}"' for inv in invalid)
            raise ValueError(f'uri contains invalid characters: {invalid}')
        return v

    class Config:
        allow_extra = True


def slugify(title):
    name = title.replace(' ', '-').lower()
    name = URI_NOT_ALLOWED.sub('', name)
    name = re.sub('-{2,}', '-', name)
    return name.strip('_-')


@contextfunction
def resolve_url(ctx, path):
    # TODO try more things, raise error on failure
    theme_files = ctx['site']['theme_files']
    return theme_files.get(path) or path



import asyncio
import contextlib
import logging
import signal
import traceback
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from time import time

from aiohttp.web_runner import AppRunner, TCPSite
from aiohttp_devtools.runserver import serve_static
from watchgod import Change, DefaultWatcher, awatch

from .assets import copy_assets, find_theme_files, run_grablib, start_webpack_watch
from .build import BuildPages, build_pages, render_pages
from .common import HarrierProblem
from .config import Config
from .data import load_data
from .extensions import apply_modifiers

HOST = '0.0.0.0'
logger = logging.getLogger('harrier.dev')


class Server:
    def __init__(self, config: Config, port: int):
        self.config = config
        self.port = port
        self.loop = asyncio.get_event_loop()
        self.runner = None

    async def start(self):
        app, *_ = serve_static(static_path=str(self.config.dist_dir), port=self.port)
        self.runner = AppRunner(app, access_log=None)
        await self.runner.setup()

        site = TCPSite(self.runner, HOST, self.port, shutdown_timeout=0.01)
        await site.start()

    async def shutdown(self):
        logger.info('shutting down server...')
        start = self.loop.time()
        with contextlib.suppress(asyncio.TimeoutError, KeyboardInterrupt):
            await self.runner.cleanup()
        logger.debug('shutdown took %0.2fs', self.loop.time() - start)


# CONFIG will be set before the fork so it can be used by the child process
CONFIG: Config = None
# SOM and BUILD_CACHE will only be set after the fork in the child process created by ProcessPoolExecutor
SOM = None
BUILD_CACHE = {}
FIRST_BUILD = '__FB__'


def update_site(pages, assets, sass, templates, extensions):  # noqa: C901 (ignore complexity)
    assert CONFIG, 'CONFIG global not set'
    start_time = time()
    global SOM
    full_build = SOM is None
    first_build = pages == FIRST_BUILD
    if first_build:
        logger.info('building...')
        full_build = True
    else:
        msg = [
            pages and f'{len(pages)} pages changed',
            assets and 'assets changed',
            sass and 'sass changed',
            templates and 'templates changed'
        ]
        logger.info('%s rebuilding...', ', '.join([m for m in msg if m]))

    log_prefix = '' if first_build else 're'
    try:
        if extensions:
            CONFIG.extensions.load()
            config = apply_modifiers(CONFIG, CONFIG.extensions.config_modifiers)
            assets = sass = templates = full_build = True
        else:
            config = CONFIG

        if assets:
            copy_assets(config)
            templates = True  # force re-render as pages might have changed
            sass = True  # in case paths changed as used by resolve_url in sass
        if sass:
            run_grablib(config)
            templates = True  # force re-render as pages might have changed

        if full_build:
            SOM = config.dict()
            SOM.update(
                theme_files=find_theme_files(config),
                pages=build_pages(config),
                data=load_data(config),
            )
            SOM = apply_modifiers(SOM, config.extensions.som_modifiers)
        elif pages:
            page_builder = BuildPages(config)
            for change, path in pages:
                rel_path = str(path.relative_to(config.pages_dir))
                if change == Change.deleted:
                    SOM['pages'][rel_path]['outfile'].unlink()
                    SOM['pages'].pop(rel_path)
                else:
                    SOM['pages'][rel_path] = page_builder.prep_file(path)
            SOM = apply_modifiers(SOM, config.extensions.som_modifiers)
            templates = templates or any(change != Change.deleted for change, _ in pages)

        if assets or sass:
            SOM['theme_files'] = find_theme_files(config)

        if templates:
            global BUILD_CACHE
            BUILD_CACHE = render_pages(config, SOM)
    except HarrierProblem as e:
        logger.debug('error during build %s %s %s', traceback.format_exc(), e.__class__.__name__, e)
        logger.warning('%sbuild failed in %0.3fs', log_prefix, time() - start_time)
        return 1
    else:
        logger.info('%sbuild completed in %0.3fs', log_prefix, time() - start_time)
        return 0


def is_within(location: Path, directory: Path):
    try:
        location.relative_to(directory)
    except ValueError:
        return False
    else:
        return True


class HarrierWatcher(DefaultWatcher):
    def __init__(self, root_path):
        self._used_paths = str(CONFIG.pages_dir), str(CONFIG.theme_dir)
        super().__init__(root_path)

    def should_watch_dir(self, entry):
        return super().should_watch_dir(entry) and entry.path.startswith(self._used_paths)


async def adev(config: Config, port: int):
    global CONFIG
    CONFIG = config
    stop_event = asyncio.Event()
    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, stop_event.set)
    loop.add_signal_handler(signal.SIGTERM, stop_event.set)

    webpack_process = await start_webpack_watch(config)

    # max_workers = 1 so the same config and som are always used to build the site
    with ProcessPoolExecutor(max_workers=1) as executor:
        ret = await loop.run_in_executor(executor, update_site, FIRST_BUILD, True, True, True, True)

        logger.info('\nStarting dev server, go to http://localhost:%s', port)
        server = Server(config, port)
        await server.start()

        try:
            async for changes in awatch(config.source_dir, stop_event=stop_event, watcher_cls=HarrierWatcher):
                logger.debug('file changes: %s', changes)
                pages, assets, sass, templates, extensions = set(), False, False, False, False
                for change, raw_path in changes:
                    path = Path(raw_path)
                    if is_within(path, config.pages_dir):
                        pages.add((change, path))
                    elif is_within(path, config.theme_dir / 'assets'):
                        assets = True
                    elif is_within(path, config.theme_dir / 'sass'):
                        sass = True
                    elif is_within(path, config.theme_dir / 'templates'):
                        templates = True
                    elif path == config.extensions.path:
                        extensions = True

                if any([pages, assets, sass, templates, extensions]):
                    ret = await loop.run_in_executor(executor, update_site, pages, assets, sass, templates, extensions)
        finally:
            if webpack_process:
                if webpack_process.returncode is None:
                    webpack_process.send_signal(signal.SIGTERM)
                elif webpack_process.returncode > 0:
                    logger.warning('webpack existed badly, returncode: %d', webpack_process.returncode)
            await server.shutdown()
        return ret



import logging
from enum import Enum
from importlib.util import module_from_spec, spec_from_file_location
from types import FunctionType

from jinja2 import (contextfilter, contextfunction, environmentfilter, environmentfunction, evalcontextfilter,
                    evalcontextfunction)

from .common import HarrierProblem, PathMatch

__all__ = (
    'modify',
    'template',
)
logger = logging.getLogger('harrier.extensions')


class ExtensionError(HarrierProblem):
    pass


class ExtType(str, Enum):
    config_modifiers = 'config_modifiers'
    som_modifiers = 'som_modifiers'
    page_modifiers = 'page_modifiers'
    template_filters = 'template_filters'
    template_functions = 'template_functions'


class Extensions:
    def __init__(self, path):
        self.path = path
        self._extensions = {}

    def __getstate__(self):
        return self.path

    def __setstate__(self, state):
        self.path = state
        self._extensions = {}

    def _set_extensions(self):
        self.config_modifiers = self._extensions[ExtType.config_modifiers]
        self.som_modifiers = self._extensions[ExtType.som_modifiers]
        self.page_modifiers = self._extensions[ExtType.page_modifiers]
        self.template_filters = self._extensions[ExtType.template_filters]
        self.template_functions = self._extensions[ExtType.template_functions]

    @classmethod
    def get_validators(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        extensions = cls(value)
        extensions.load()
        return extensions

    def load(self):
        self._extensions = {
            ExtType.config_modifiers: [],
            ExtType.som_modifiers: [],
            ExtType.page_modifiers: [],
            ExtType.template_filters: {},
            ExtType.template_functions: {},
        }
        if self.path.exists():
            spec = spec_from_file_location('extensions', self.path)
            module = module_from_spec(spec)
            try:
                spec.loader.exec_module(module)
            except Exception as e:
                logger.exception('error loading extensions %s %s', e.__class__.__name__, e)
                raise ExtensionError(str(e)) from e

            for attr_name in dir(module):
                if attr_name.startswith('_'):
                    continue
                attr = getattr(module, attr_name)
                ext_type = getattr(attr, '__extension__', None)
                if ext_type == ExtType.page_modifiers:
                    self._extensions[ext_type].extend([(path_match, attr) for path_match in attr.path_matches])
                elif ext_type:
                    self._extensions[ext_type].append(attr)
                elif any(getattr(attr, n, False) is True for n in filter_attrs):
                    self._extensions[ExtType.template_filters][attr_name] = attr
                elif any(getattr(attr, n, False) is True for n in function_attrs):
                    self._extensions[ExtType.template_functions][attr_name] = attr
        self._set_extensions()

    def __repr__(self):
        ext = self._extensions and {k.value: v for k, v in self._extensions.items()}
        return f'<Extensions {repr(ext) if ext else "not loaded"}>'

    def __eq__(self, other):
        return self._extensions == getattr(other, '_extensions', other)


def apply_modifiers(obj, ext):
    original_type = type(obj)
    for f in ext:
        try:
            obj = f(obj)
        except Exception as e:
            logger.exception('error running extension %s', f.__name__)
            raise ExtensionError(str(e)) from e

        if not isinstance(obj, original_type):
            logger.error('extension "%s" did not return a %s object as expected', f.__name__, original_type.__name__)
            raise ExtensionError(f'extension "{f.__name__}" did not return a {original_type.__name__} '
                                 f'object as expected')
    return obj


class modify:
    @staticmethod
    def config(f):
        f.__extension__ = ExtType.config_modifiers
        return f

    @staticmethod
    def som(f):
        f.__extension__ = ExtType.som_modifiers
        return f

    @staticmethod
    def pages(*globs):
        if not globs:
            raise HarrierProblem('validator with no page globs specified')
        elif isinstance(globs[0], FunctionType):
            raise HarrierProblem("modify_pages should be used with page globs as arguments, not bare. "
                                 "E.g. usage should be `@modify_pages('<page_glob>', ...)`")

        def dec(f):
            f.__extension__ = ExtType.page_modifiers
            f.path_matches = [PathMatch(glob) for glob in globs]
            return f
        return dec


filter_attrs = 'contextfilter', 'evalcontextfilter', 'environmentfilter', '__vanilla_filter__'
function_attrs = 'contextfunction', 'evalcontextfunction', 'environmentfunction', '__vanilla_function__'


class template:
    contextfilter = contextfilter
    evalcontextfilter = evalcontextfilter
    environmentfilter = environmentfilter
    contextfunction = contextfunction
    environmentfunction = environmentfunction
    evalcontextfunction = evalcontextfunction

    @staticmethod
    def filter(f):
        f.__vanilla_filter__ = True
        return f

    @staticmethod
    def function(f):
        f.__vanilla_function__ = True
        return f



from importlib.machinery import SourceFileLoader
from pathlib import Path
from setuptools import setup

description = 'Static site generator'
THIS_DIR = Path(__file__).resolve().parent
try:
    long_description = '\n\n'.join([
        THIS_DIR.joinpath('README.rst').read_text(),
        # THIS_DIR.joinpath('HISTORY.rst').read_text()
    ])
except FileNotFoundError:
    long_description = description + '.\n\nSee https://TODO.com/ for documentation.'

# avoid loading the package before requirements are installed:
version = SourceFileLoader('version', 'harrier/version.py').load_module()

setup(
    name='harrier',
    version=str(version.VERSION),
    description=description,
    long_description=long_description,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='css,sass,scss,jinja,jinja2,build,static,static site generator',
    author='Samuel Colvin',
    author_email='s@muelcolvin.com',
    url='https://github.com/samuelcolvin/harrier',
    license='MIT',
    packages=['harrier'],
    python_requires='>=3.6',
    zip_safe=True,
    entry_points="""
        [console_scripts]
        harrier=harrier.cli:cli
    """,
    install_requires=[
        'Jinja2>=2.8',
        'Pygments>=2.2.0',
        'aiohttp-devtools>=0.9',
        'click>=6.2',
        'devtools>=0.3',
        'grablib>=0.7.3',
        'libsass>=0.10.1',
        'misaka>=2.1.0',
        'pydantic>=0.8',
        'ruamel.yaml>=0.15.37',
        'watchgod>=0.1.1',
    ]
)
